TOKEN_TEMPLATE = (
    f'https://login.microsoftonline.com/{{tenant_id}}/oauth2/v2.0/token'
)
FABRIC_SCOPE = 'https://api.fabric.microsoft.com/.default'
POWERBI_SCOPE = 'https://analysis.windows.net/powerbi/api/.default'
POWERBI_API = 'https://api.powerbi.com/v1.0/myorg'
FABRIC_API = 'https://api.fabric.microsoft.com/v1'
PLATFORM_SCHEMA = 'https://developer.microsoft.com/json-schemas/fabric/gitIntegration/platformProperties/2.0.0/schema.json'
PLATFORM_VERSION = '2.0'
