_FABRIC_ITEMS = {
    'ApacheAirflowJob': 'apache_airflow_jobs',
    'CopyJob': 'copy_jobs',
    'Dashboard': 'dashboards',
    'DataPipeline': 'data_pipelines',
    'Dataflow': 'dataflows_gen2',
    'Datamart': 'datamarts',
    'DigitalTwinBuilder': 'digital_twin_builders',
    'DigitalTwinBuilderFlow': 'digital_twin_builder_flows',
    'Environment': 'environments',
    'Eventhouse': 'eventhouses',
    'Eventstream': 'eventstreams',
    'GraphQLApi': 'graphql_apis',
    'KQLDashboard': 'kql_dashboards',
    'KQLDatabase': 'kql_databases',
    'KQLQueryset': 'kql_querysets',
    'Lakehouse': 'lakehouses',
    'MLExperiment': 'ml_experiments',
    'MLModel': 'ml_models',
    'MirroredAzureDatabricksCatalog': 'mirrored_azure_databricks_catalogs',
    'MirroredDatabase': 'mirrored_databases',
    'MirroredWarehouse': 'mirrored_warehouses',
    'MountedDataFactory': 'mounted_data_factories',
    'Notebook': 'notebooks',
    'PaginatedReport': 'paginated_reports',
    'Reflex': 'reflexes',
    'Report': 'reports',
    'SQLDatabase': 'sql_databases',
    'SQLEndpoint': 'sql_endpoints',
    'SemanticModel': 'semantic_models',
    'SparkJobDefinition': 'spark_job_definitions',
    'VariableLibrary': 'variable_libraries',
    'Warehouse': 'warehouses',
    'WarehouseSnapshot': 'warehouse_snapshots',
}
