shortcuts_payloads_targets = {
    'adlsGen2': {
        'connectionId': '{{target_connection_id}}',
        'location': '{{target_location}}',
        'subpath': '{{target_subpath}}',
    },
    'amazonS3': {
        'connectionId': '{{target_connection_id}}',
        'location': '{{target_location}}',
        'subpath': '{{target_subpath}}',
    },
    'azureBlobStorage': {
        'connectionId': '{{target_connection_id}}',
        'location': '{{target_location}}',
        'subpath': '{{target_subpath}}',
    },
    'dataverse': {
        'connectionId': '{{target_connection_id}}',
        'deltaLakeFolder': '{{target_delta_lake_folder}}',
        'environmentDomain': '{{target_environment_domain}}',
        'tableName': '{{target_table_name}}',
    },
    'googleCloudStorage': {
        'connectionId': '{{target_connection_id}}',
        'location': '{{target_location}}',
        'subpath': '{{target_subpath}}',
    },
    'oneLake': {
        'itemId': '{{target_item_id}}',
        'path': '{{target_path}}',
        'workspace_id': '{{target_workspace_id}}',
        'connectionId': '{{target_connection_id}}',
    },
    's3Compatible': {
        'bucket': '{{target_bucket}}',
        'connectionId': '{{target_connection_id}}',
        'location': '{{target_location}}',
        'subpath': '{{target_subpath}}',
    },
}
